;----------------------------------------------------------------------------;
; Capacitance Meter Control Program R0.01  (C)ChaN, 2003
;
; 
;
;----------------------------------------------------------------------------;

.include "tn2313def.inc"		;Device definition file included in "AVR Family Assembler".
.include "avr.inc"

.def	Zero	= r15			;Zero register
.def	Stm1	= r14			;System timer (250Hz decrement/zero stopped)
.def	Stm2	= r13			;System timer (250Hz decrement/zero stopped)

.def	Flags	= r25			;b0:Result is minus
								;b1:Capture completed
								;b2:Integration time out
								;b7:Button pressed


;----------------------------------------------------------;
; Eeprom area
.eseg	
	
	;.db	0xC8,0x97,0xB5,0x95
	.db 0xC8,0xBB,0x97,0x52,0x96,0x8F

;----------------------------------------------------------;
; Data memory area

.dseg
	.org	SRAM_START

	DispPtr: .byte	1	;Display buffer
	DispBuf: .byte	6	;
	KeyScan: .byte	2	;

	Comp1:	 .byte	2	;Range low compensation
	Comp2:	 .byte	2	;Range high compensation
	Comp3:	 .byte	2	;Zero compensation value

	StrBuf:	 .byte	10	;Decimal conversion buffer


;----------------------------------------------------------;
; Program code area

.cseg
	rjmp	reset		;Reset
	rjmp	0			;External INT0
	rjmp	0			;External INT1
	rjmp	tc1_cap		;TC1 capture
	rjmp	0			;TC1 compare
	rjmp	tc1_ovf		;TC1 overflow
	rjmp	tc0_ovf		;TC0 overflow


;----------------------------------------------------------;
; Initialize

reset:
	ldi		AL, LOW(RAMEND)
	out		SPL,AL
	clr		Zero					;Permanent zero reg.

	ldi		ZL,LOW(SRAM_START)
	ldi		ZH,HIGH(SRAM_START)

	ldi		AL, 128			
	st		Z+, Zero			
	dec		AL			
	brne	PC-2			
	
	ser		AL
	out		DDRD,  AL				; PinD0 bis PinD6 auf Ausgang

	ldi		AL, 0b01111000			; Pullup aktivieren fr PortB3, Taste(4),PortB5, Miso(6)
	out		PORTB, AL
	ldi		AL, 0b10101101			; Ausgang B0,B2,B3,B5,B7
	out		DDRB, AL

	ldi		AL, (1<<CS02)
	out		TCCR0B, AL				; ClK/256 => TC0 = 39kHz

	ldi		AL, (1<<TOIE0)
	out		TIMSK, AL				

	clr		Flags					; Flags lschen
	rcall 	lcd_init				; LCD initialisieren
   	rcall	lcd_clear				; Display lschen (notwendig!!!)

	sbi		ACSR, ACIC				; Analog Comparator, AC Input Capture

	sei

	rcall	load_eep				; Kalibrierungswerte laden
	breq	PC+6					; wenn vorhanden, weiter zu main
	
    ldi		ZL,low(form3*2)			; sonst Eeprom-Fehler ausgeben
	ldi		ZH,high(form3*2)
	rcall	lcd_print				; warten
	ldi		AL, 250					;
	rcall	dly						;
	rcall	lcd_clear				; LCD wieder lschen

	rjmp	can_offset

;----------------------------------------------------------;
; Command processing loop (main)

main:
	ldi		AL, 25					; Wait for 100ms and Timer2 erapsed.
	mov		Stm1, AL		
	cbr		Flags, bit7		
	sbrc	Flags, 7		
	rjmp	btn_pressed
	tst		Stm1			
	brne	PC-3			
	tst		Stm2			
	brne	PC-5			

	ldi		AL, 125					; Start Timer2 (500ms)
	mov		Stm2, AL		
	rcall	measure					; Messen im niedrigen Bereich
	brcc	m1						; Bei 'time out', mittl. Bereich
	cbi		PORTB,3					; lsche PINB3
	rcall	measure					; messen im mittleren Bereich
	brcc	m1						; 
	sbi		PortB,3
	cbi		PortB,5					; messen im oberen Bereich
	rcall	measure					
m1: rcall	adjust_zero				 
	rcall	adjust_gain				
	rcall	disp_val				
	rcall	disp_measure
	sbi		PORTB,3					; zurck zu Low-Range
	sbi		PortB,5	

	rjmp	main

btn_pressed:
	ldi		AL, 4					; Delay 16ms
	rcall	dly				
	sbis	PINB, 6					; Ist ISP 9/10 berbrckt?
	rjmp	cal_low					; Ja=> Kalibrierung Low-Range
	;sbis	PINB, 5					; Ist ISP 7/8 berbrckt?
	sbis	PinB, 7
	rjmp	cal_high		        ; JA=> Kalibrierung Mid-Range
	rjmp	can_offset				; sonst Nullpunkt setzen


cal_high:
	cbi		PORTB, 3				; Mid-Range setzen
	rcall	measure					; messen
	sbi		PORTB, 3		
	ldiw	C, 0					; X:D:C = 1000*65536; (100nF reference cap)
	ldiw	D, 1000			
	ldiw	X, 0			
	ldiw	Z, Comp2				; Speicherstelle holen
	rjmp	cal_comp

cal_low:
	rcall	measure					; messen
	rcall	adjust_zero				
	ldiw	C, 0					; X:D:C = 10000*65536; (1nF reference cap)
	ldiw	D, 10000		
	ldiw	X, 0			
	ldiw	Z, Comp1

cal_comp:
	clrw	T0						; X:D:C /= B:A;
	clrw	T2				
	ldi		EL, 48			
	lslw	C				
	rolw	D				
	rolw	X				
	rolw	T0				
	rolw	T2				
	cpw		T0, A			
	cpcw	T2, B			
	brcs	PC+6			
	subw	T0, A			
	sbcw	T2, B			
	inc		CL				
	dec		EL					
	brne	PC-21			
	or		DL, DH					; Check over flow
	or		DL, XL				
	or		DL, XH				
	brne	cal_err					
	stdw	Z+0, C
  	rcall	clr_disp
	rcall	save_eep
	rjmp	main

can_offset:
	rcall	measure					; Measure capacitance as zero
	or		BL, BH					; Check adjustment range
	brne	cal_err					;
	cpi		BH, high(2000)			;
	brcc	cal_err					; 
	stsw	Comp3, A				; Set the value as zero point
	rcall	clr_disp
	rjmp	main

cal_err:
	rcall	lcd_home
	ldiw	Z, form4*2				; Fehler ausgeben
	rcall 	lcd_print
	ldi		AL, 250
	rcall	dly
	rcall	lcd_clear
	rjmp	main

dly:
	mov		Stm1, AL
	tst		Stm1	
	brne	PC-1
	ret


;----------------------------------------------------------;
; Measure capacitance

measure:
	out		TCNT1H, Zero				; Clear TC1 and set time limit
	out		TCNT1L, Zero				;
	clr		T2L							;
	ldi		AL, 20						;
	
	sbis	PORTB,3						; berspringe, wenn PortB3 gesetzt
	ldi		AL, 152
	sbis	Portb,5						; berspringe, wenn PortB5 gesetzt
	ldi		AL, 152						;
	
	mov		T2H, AL						;
	ldi		AL, (1<<TOV1)|(1<<ICF1)
	out		TIFR, AL					
	ldi		AL, (1<<TOIE1)|(1<<ICIE1)|(1<<TOIE0)
	out		TIMSK,AL					; T/C Overflow Interrupt 1, ICE1, Intrpt 0
	cbr		Flags, bit0+bit1+bit2		;
	ldi		AL, (1<<ICES1)|(1<<CS10)
	out		TCCR1B, AL					; ICES1 CS10 Timer Capture / no prescaling
	cbi		DDRB, 0						; Start to charge

	sbrc	Flags, 2					; Wait for end of integration
	rjmp	mea_over					;
	sbrs	Flags, 1					;
	rjmp	PC-3						;

	ldi		AL, (1<<ICES1)
	out		TCCR1B, AL					; Stop TC1
	movew	A, T4						; Ergebnis in AH:AL
	movew	B, T6						; und BH:BL
	clc
	ret

mea_over:
	ldi		AL, (1<<ICES1)
	out		TCCR1B, AL					; Stop TC1
	ldi		AL, 4						; Wait for 16ms
	rcall	dly							;
	ldi		BH, -1
	sec                        			; set Carry-Flag
	ret


adjust_zero:
	sbis	PORTB, 3					; Skip if in mid range
	rjmp	az0
	sbis	Portb, 5
	rjmp	az0							; 
	ldsw	C, Comp3					; B:A -= Comp3;
	subw	A, C						;
	sbc		BL, Zero					
	sbc		BH, Zero					
	brcc	az0;PC+10					; if sign, B:A *= -1; and set sign flag.
	sbr		Flags, bit0					
	comw	A							
	comw	B							
	adc		AL, Zero					;
	adc		AH, Zero					;
	adc		BL, Zero					;
	adc		BH, Zero					;
az0:
	ret

adjust_gain:
	ldiw	Y, Comp1
	sbis	PORTB,3						; Gain adjustment
	adiw	YL,2
	sbis	PortB,5
	adiw	YL, 2						; Load compensation value in to D by range
	lddw	D, Y+0						; 
	subw	C, C						; B:A = B:A * D / 65536;
	ldi		EL, 33						;
ag0:
	brcc	ag1
	addw	C, D						;
ag1:
	rorw	C							;
	rorw	B							;
	rorw	A							;
	dec		EL							;
	brne	ag0
	movew	A, B						;
	movew	B, C						;
	ret



;----------------------------------------------------------;
; Ergebnis in B:A in 0.1pF

disp_val:
	ldiw	X, StrBuf					; Puffer
	clr		DL							; Anzahl Stellen
dv1:	
	inc		DL							; Zhler1 = Zhler1 + 1
	clr		CL							; Zhler2 lschen
	ldi		CH,32						; Zhler2=32 	
dv0:
	lslw	A							; A
	rolw	B							; und B
	rol		CL							; und CL * 2
	cpi		CL,10						; wenn CL<10 weiter mit schieben
	brcs	dv3							; 	
	subi	CL,10						; sonst CL=CL-1
	inc		AL							; Rest unten nachschieben
dv3:
	dec		CH							; Zhler2=Zhler2-1
	brne	dv0							; zurck zum Anfang	
	st		X+, CL						; Wert in CL sichern
	cp		AL, Zero					; ist Al	
	cpc		AH, Zero					; und Ah	
	cpc		BL, Zero					; und Bl
	cpc		BH, Zero					; und BH =0
	brne	dv1							; wenn nicht zurck
	cpi		DL, 2						; ist Zhler2 = 2 Stellen
	brcc	dv2							; wenn ja weiter
	st		X+, Zero					; wenn nicht
	inc		DL							; Zhler1 erhhen
dv2:	
	sbis	PORTB, 3					; Ist  PortB gelscht?
	addi	DL,3
	sbis	PortB, 5
	addi	DL, 4						; wenn ja, Zhler2=Zhler2+3
	ldiw	Z, form2*2-6		 
	sbrs	Flags, 0					; sind Flags auf 0
	adiw	ZL, 24						; wenn ja 4 Masken addieren
dv4:
	adiw	ZL, 6						; so viele Masken
	dec		DL							; wie Stellen
	cpi		DL, 2						; addieren
	brcc	dv4

put_formed:
	clr		AH
	ldi		YL, Low(DispBuf)
	ldi		YH, high(DispBuf)
pf0:
	lpm
	adiw	ZL, 1
	mov		AL, T0L
	clt
	cpi		AL,0
	brne	pf2
	bst		AL, 0
	ld		AL, -X
pf2:
	pushw	Z
	ldiw	Z, seg7*2
	addw	Z,A
	lpm
	popw	Z
	st		Y+, T0L
	cpi		YL, DispBuf+6
	brne	pf0
	ret


clr_disp:
	
	ldiw	Y,  DispBuf
	st		Y+, Zero
	cpi		YL, DispBuf+6
	brne	PC-2
	ret


disp_measure:
	cli
	rcall	lcd_Home
	ldiw	Z,DispBuf
d0:	ld		AL,Z+	
	rcall	lcd_Data					; Ausgabe
	cpi		ZL,DispBuf+6				; ist Zhler=0 ?
	brne	d0                			; wenn nicht, nchstes Zeichen
	sei
	ret



form4:	.db "Kalibrierung!  ",0			

form3:	.db	"EEPROM-Fehler! ",0

form2:	.db 10,0,16,0,13,17 	;-0,0pF
		.db 15,10,0,0,13,17		; -00pF
		.db 15,14,3,15,15,15 	; E3

		.db 15,14,3,15,15,15	; E3
		.db 0,16,0,0,13,17		;0,00pF
		.db 0,0,16,0,13,17   	;00,0pF
		.db 15,0,0,0,13,17		; 000pF
		.db 0,16,0,0,12,17		;0,00nF
		.db 0,0,16,0,12,17		;00,0nF
		.db 15,0,0,0,12,17   	; 000nF
		.db 0,16,0,0,11,17		;0,00uF
		.db 0,0,16,0,11,17		;00,0uF
		.db 15,0,0,0,11,17		; 000uF
		.db 0,0,0,0,11,17		;0000uF
		;.db 0,0,0,0,11,17		;0000uF
		;.db 0,0,0,0,11,17		;0000uF
		.db 15,19,18,15,15,15 	; 0L
		.db 15,19,18,15,15,15 	; 0L
		.db 15,19,18,15,15,15 	; 0L

seg7:	.db '0','1','2','3','4','5','6','7','8','9'
		.db '-',0xE4,'n',0xE6,'E',' ',',','F','L','O'
		;    10,  11, 12, 13, 14, 15, 16, 17, 18, 19	


;----------------------------------------------------------;
; Load/Save EEPROM

load_eep:
	ldiw	Y, Comp1					; Load compensation data
	ldiw	C, 0x5501	
	rcall	read_eep	
	st		Y+, AL		
	add		CH, AL		
	cpi		YL, Comp1+4	
	brne	PC-4		
	rcall	read_eep					; Check SUM
	cp		AL, CH		
	breq	PC+6		
	ldi		AL,-1
	st		-Y, AL						; Set default value if data have been broken.
	st		-Y, AL		
	st		-Y, AL		
	st		-Y, AL		
	ret


save_eep:
	ldiw	Y, Comp1					; Save compensation data
	ldiw	C, 0x5501		
	ld		AL, Y+		
	add		CH, AL		
	rcall	write_eep	
	cpi		YL, Comp1+4	
	brne	PC-4		
	mov		AL, CH						; Save check SUM

write_eep:
	out		EEAR, CL
	inc		CL
	out		EEDR, AL
	cli
	sbi		EECR, EEMPE
	sbi		EECR, EEPE
	sei
	sbic	EECR, EEPE
	rjmp	PC-1
	ret

read_eep:
	out		EEAR, CL
	inc		CL
	sbi		EECR, EERE
	in		AL, EEDR
	ret


;----------------------------------------------------------;
; TC1 overflow interrupt
;
; T2L counts carry outs from TCNT1. When T2L reaches T2H,
; a time-out error flag will be set.

tc1_ovf:
	push	AL
	in		AL, SREG
	push	AL

	inc		T2L
	cp		T2L, T2H
	brcs	PC+6
	sbi		DDRB, 2
	sbi		DDRB, 0
	sbr		Flags, bit2
	ldi		AL, (1<<TOIE0)
	out		TIMSK, AL

	pop		AL
	out		SREG, AL
	pop		AL
	reti



;----------------------------------------------------------;
; TC1 capture interrupt
;
; When Vc reaches 0.17 Vcc, capture t1 and change reference
; voltage to 0.5Vcc. When Vc reaches 0.5 Vcc, capture t2 and
; terminate the measureing.

tc1_cap:
	push	AL
	in		AL, SREG
	push	AL
	sbis	DDRB, 2							; Branch by measureing phase.
	rjmp	tc1c_ed				 

tc1c_st:									; Vc reaches 0.17 Vcc
	in		T4L, ICR1L						; Capture t1
	in		T4H, ICR1H			
	mov		T6L, T2L			 
	cbi		DDRB, 2							; Change Vth to 0.5 Vcc.
	ldi		AL, 20							; Deley several microseconds and clear Irq.
	dec		AL					
	brne	PC-1				
	ldi		AL, (1<<ICF1)
	out		TIFR, AL
	rjmp	tc1c_e

tc1c_ed:									; Vc reaches 0.5 Vcc
	mov		T6H, T4L						; Capture t2-t1
	in		T4L, ICR1L			
	sub		T4L, T6H			
	mov		T6H, T4H			
	in		T4H, ICR1H			
	sbc		T4H, T6H			
	mov		T6H, T6L			
	mov		T6L, T2L			
	sbc		T6L, T6H			
	clr		T6H					
	;sbi		DDRB, 2						; Set Vth to 0.17 Vcc.
	outi	TIMSK, 0b00000010				; Disable Irq.
	sbi		DDRB, 0							; Discharge capacitor
		
	sbic	PortB, 3						; wenn Mid-Range nicht gesetzt
	rjmp	tc1								; weiter bei tc1
	rcall	wait100ms						; 100ms entladen
	rjmp	tc2
tc1:
	sbic	PortB,5 						; wenn High-Range nicht gesetzt
	rjmp	tc2								; weiter zu Ende
	rcall	wait500ms						; 500 ms entladen
	
tc2:
	sbi		DDRB, 2							; Set Vth to 0.17 Vcc.
	sbr		Flags, bit1						; Flag: Messung-Ende

tc1c_e:
	
	pop		AL
	out		SREG, AL
	pop		AL
	reti


;----------------------------------------------------------;
; TC0 overflow interrupt (1kHz)
;
; - Refresh LED display.
; - Scan button inputs.
; - Decrement Stm1 and Stm2. (250Hz)

tc0_ovf:
	push	AL					
	outi	TCNT0, -39
	sei
	in		AL, SREG
	pushw	A
	pushw	Z

	ldi		ZL, low(DispPtr)				
	ldi		ZH, high(DispPtr)	
	ld		AH, Z							
	inc		AH								
	cpi		AH, 4							
	brcs	t0								
	rcall	scan_key			
	clr		AH
t0:	st		Z+, AH								
	popw	Z
	popw	A
	out		SREG, AL
	pop		AL
	reti

scan_key:
	in		AL, PINB						
	com		AL									
	andi	AL, bit4	      					
	ldd		AH, Z+6		   						
	std		Z+6, AL		
	cp		AL, AH		
	brne	PC+7			
	ldd		AH, Z+5		
	std		Z+5, AL
	eor		AH, AL		
	and		AH, AL		
	breq	PC+2			
	sbr		Flags, bit7							; Flag 'Button_pressed' setzen

	tst		Stm1								; Decrement Stm with zero stopeed...
	breq	PC+2			
	dec		Stm1		
	tst		Stm2		
	breq	PC+2		
	dec		Stm2		

	ret


;***************************************************************************
;*
;* LCD-Routinen  4-Bit
;*                                             	
;* PIN_LCD_Enable	= PD5									
;* PIN_LCD_RS		= PD4									
;* PIN_LCD_D4		= PD0									
;* PIN_LCD_D5		= PD1									
;* PIN_LCD_D6 		= PD2									
;* PIN_LCD_D7		= PD3									
;*																
;****************************************************************************


;sendet Datenbyte an das LCD
lcd_data:
   	
	push	AL
   	swap  	AL				   	   				; Obere 4 bits (Nibble) nach unten
  	andi  	AL, 0b00001111         				; 0bere Hlfte auf Null setzen
	sbr		AL, (1<<PIN_LCD_RS)					; PIN_LCD_RS						
	out		PORTD,AL
	rcall 	lcd_enable             				; Enable-Routine aufrufen
   	pop		AL									; 2. Nibble	
	andi  	AL, 0b00001111         				; obere Hlfte auf Null setzen
	sbr		AL, (1<<PIN_LCD_RS)					; PIN_LCD_RS
	out		PORTD,AL
	rcall 	lcd_enable             				; Enable-Routine aufrufen
	rcall 	wait50us              				; warte
	ret                            				; zurck zum Hauptprogramm

 ;sendet einen Befehl an das LCD
lcd_command:                            		; wie lcd_data, nur ohne RS zu setzen
    		
	cbi		PORTD, PIN_LCD_RS
	push	AL
   	swap	AL
	andi	AL, 0b00001111
	out		PORTD,AL
	rcall	lcd_enable
   	pop		AL
	andi	AL, 0b00001111
	out		PORTD,AL
	rcall	lcd_enable
   	rcall	wait50us
   	ret


 ;erzeugt den Enable-Puls
lcd_enable:
 	sbi 	PORTD, PIN_LCD_Enable    			; Enable high
   	nop                          				; 3 Taktzyklen warten
	nop
	nop
	cbi		PORTD, PIN_LCD_Enable     			; Enable wieder low
	ret                          				; und wieder zurck


 ;Initialisierung
lcd_init:
   	rcall  	wait1s

	ldi		AL, 0b00000011         		
   	out		PORTD, AL              		
	rcall	lcd_enable             				; 1. muss 3mal hintereinander gesendet
   	rcall	wait5ms								;    werden zur Initialisierung
   	rcall	lcd_enable             				; 2.
   	rcall	wait5ms
   	rcall	lcd_enable             				; und 3!
   	rcall	wait5ms
	
	ldi		AL,	0b00000010						; 4bit Modus einstellen
	out		PORTD,AL
	rcall	lcd_enable
	rcall	Wait5ms	

	ldi		AL, 0b00101000         				; 4bit-Modus, 2 Zeilen
   	rcall	lcd_command							; Befehl senden
	ldi		AL, 0b00001100         				; LCD on
   	rcall	lcd_command							; Befehl senden
	ldi		AL, 0b00000100         				; Entry mode
   	rcall	lcd_command
   	ret

 ;Sendet den Befehl zur Lschung des Displays
lcd_clear:
   	ldi		AL, 0b00000001         				; Display lschen
   	rcall	lcd_command
   	rcall	Wait5ms
   	ret

 ;Setzt Cursor-Position auf Anfang
lcd_home:
	ldi		AL, 0b00000010         				; Cursor an Anfang
  	rcall	lcd_command
  	rcall	wait5ms
  	ret	

; Erwartet die Position in AL			
lcd_cursor:
	sbr		AL, 0b10000000         				; Bit 7 setzen
	rcall	lcd_command
	rcall	wait5ms
   	ret

lcd_print:
	cli
	lpm											; Laden
	tst		R0									; ist Zeichen =0
	breq	lcd_pend							; gehe zu end
	mov		AL,R0								; sonst Zeichen nach AL
	rcall	lcd_data							; Ausgabe
	adiw	ZL,1              					; Zeiger in ZL = ZL + 1(word)
	rjmp	lcd_print				
lcd_pend:
 	sei
	ret



;***************************************************************************
;*
;* Warteschleifen
;*                                             	
;****************************************************************************


wait500ms:
	push		AL
	ldi			AL,5
	rjmp		n1	
wait1s: ;1005,55
	push		AL
	ldi			AL,10
n1:	rcall		wait100ms
	dec			AL
	brne		n1
	pop			AL
	ret

wait100ms:; 100,55ms
	push		AL
	ldi			AL,10
n2:	rcall		wait10ms
	dec			AL
	brne		n2
	pop			AL
	ret

wait5ms:; 5,03ms
	push		AL
	ldi			AL,5
	rjmp		n3
wait10ms: ;10,06ms
	push		AL
	ldi			AL,10
n3:	rcall		wait1ms
	dec			AL
	brne		n3
	pop			AL
	ret

wait1ms:		;1,01ms
	push		AL
	ldi			AL,20		
n4:	rcall		wait50us
	dec			AL
	brne		n4
	pop			AL
	ret

wait50us:		;49,9us
	push		AL
	ldi			AL,161		
n5:	dec			AL
	brne		n5
	pop			AL
	ret

